"""Console script for serveliza."""
import argparse
import sys
from . import __version__, __author__
from serveliza.roll.exporter import RollExporter
from serveliza.roll import ElectoralRoll
from serveliza import serveliza

DESC = 'Serveliza is an application to extract data of ' \
       'the Chilean Electoral Service (SERVEL) from different ' \
       'sources.'

DESC_SUBCMDS = 'Serveliza has different utilities to extract data ' \
               'which are accessed through its subcommand. For more ' \
               'information check the help of each one.'

DESC_ROLL = 'The roll command allows the extraction of ' \
            'electoral roll data from pdf files to csv files.'

EPILOG = f'Made with ♥ by @{__author__}.'


def roll_cli_wrapper(args, parser):
    source = args.source
    if not source:
        parser.print_help()
        return 0
    kwargs = {
        'source': source,
        'output': args.output,
        'processor': args.processor,
        'mode': args.mode,
        'mode_sep': args.separator,
        'no_suffix': args.no_suffix,
        'recursive': args.recursive,
        'no_summary': args.no_summary,
        'silent': args.silent,
        'no_colors': args.no_colors}
    try:
        serveliza.roll_from_pdf_to_csv(**kwargs)
    except TypeError as error:
        print(f'Error! > {error}')


def roll_parser(subparser):
    parser_roll = subparser.add_parser(
        'roll', help=DESC_ROLL, description=DESC+' '+DESC_ROLL,
        epilog=EPILOG)
    parser_roll.set_defaults(func=roll_cli_wrapper)
    parser_roll.add_argument(
        'source', nargs='*', type=str,
        help=ElectoralRoll.source.__doc__)
    parser_roll.add_argument(
        '-o', '--output', help=RollExporter.output.__doc__,
        type=str, metavar='output', default='output')
    processors = [x[0] for x in ElectoralRoll.processor_ref.items()]
    parser_roll.add_argument(
        '-p', '--processor', help=ElectoralRoll.processor.__doc__,
        type=str, default='pdftotext',
        choices=processors)
    parser_roll.add_argument(
        '-m', '--mode', help=RollExporter.mode.__doc__,
        type=str, default='unified', choices=RollExporter.modes)
    parser_roll.add_argument(
        '-s', '--separator', help=RollExporter.mode_sep.__doc__,
        type=str, choices=RollExporter.mode_sep_opts)
    parser_roll.add_argument(
        '-r', '--recursive', help=ElectoralRoll.recursive.__doc__,
        action='store_true', default=False)
    parser_roll.add_argument(
        '--no-suffix', help=RollExporter.random_suffix.__doc__,
        action='store_true', default=False)
    parser_roll.add_argument(
        '--no-summary', help=RollExporter.summary.__doc__,
        action='store_true', default=False)
    parser_roll.add_argument(
        '--silent', help='Does not print application progress on screen.',
        action='store_true', default=False)
    parser_roll.add_argument(
        '--no-colors', help='Does not colorize screen prints.',
        action='store_true', default=False)
    return parser_roll


def main():
    '''Console script for serveliza.'''
    parser = argparse.ArgumentParser(
        prog='serveliza', description=DESC, epilog=EPILOG)
    # version
    parser.add_argument(
        '-v', '--version', action='version',
        version="%(prog)s "+__version__)
    subparser = parser.add_subparsers(
        title='sub-commands', description=DESC_SUBCMDS, help='description:')
    # roll subcommand parser:
    parser_roll = roll_parser(subparser)
    # insert other subcommands here:
    # parser_cmd = cmd_parser(subparser)
    # ...
    args = parser.parse_args()
    if hasattr(args, 'func') and args.func == roll_cli_wrapper:
        args.func(args, parser_roll)
    else:
        parser.print_help()
    return 0


if __name__ == '__main__':
    sys.exit(main())  # pragma: no cover



from datetime import datetime as dt
from datetime import timedelta
from pathlib import Path
from pandas import pandas as pd
import numpy as np


from serveliza.mixins.pdf import PDFProcessorMixin
from serveliza.utils import pdf as pdf_utils
from .parsers import RollParser
from .adapters import RollAdapter
from .printer import RollPrinter
from .memorizer import RollMemorizer
from .exporter import RollExporter


DURATIONS_SCHEMA = {
    'processing': timedelta(),
    'adapting': timedelta(),
    'parsing': timedelta(),
    'memorizing': timedelta(),
    'exporting': timedelta()}


class ElectoralRoll(PDFProcessorMixin):
    '''ElectoralRoll

    :class:`ElectoralRoll <.ElectoralRoll>` allows to \
    instantiate an electoral roll of the chilean Electoral Service \
    (*Servicio Electoral de Chile*, SERVEL) from PDF files.

    Different parameters are handled in the constructor of this class. In \
    itself it handles the *source* parameter that determines the path \
    where to recognize pdf files (it can be a directory path or a file), \
    the *recursive* parameter that determines whether to search in the \
    root of the defined path or in each of its subdirectories, and \
    the *auto* parameter that determines if the extraction is automatic. \
    It also inherits parameters from :class:`PDFProcessorMixin \
    <.PDFProcessorMixin>` (*processor*). Likewise, in the constructor \
    it instantiates other nested classes by routing their parameters.

    :param str source: The source path of pdf files.
    :param bool auto: Run the extract in the instantation.
    :param bool recursive: Determines if the search for pdf files in the \
        delivered source is recursive or is only for the root of the \
        indicated directory,
    :param str processor: Processor to use (default='pdftotext', see \
        more in :class:`PDFProcessorMixin <.PDFProcessorMixin>`).
    :param bool memorize: Storage data in memory of instance (default=True, \
        see more in :class:`RollMemorizer <.RollMemorizer>`).
    :param bool export: If export data in csv file (default=False, \
        see more in :class:`RollExporter <.RollExporter>`).
    :param str output: Directory to store the data in csv file(s) (\
        see more in :class:`RollExporter <.RollExporter>`).
    :param str mode: Determines the data export mode in files. If it \
        is *unified* (default) it creates a single csv file with the data,\
         or if it is *separated* into several according to communal or \
        regional criteria (see more in :class:`RollExporter <.RollExporter>`).
    :param str mode_sep: Criteria for separating files in export in \
        separate mode (*commune* or *region*, default="commune", \
        see more in :class:`RollExporter <.RollExporter>`).
    :param bool random_suffix: Determines whether exported files have a \
        random text string appended to the end (see more in \
        :class:`RollExporter <.RollExporter>`).
    :param bool summary: Determines whether to generate a summary file of \
        the export and the extracted data (see more in :class:`RollExporter \
        <.RollExporter>`).

    Anyway, only the *source* parameter is required:

    >>> obj = ElectoralRoll(source='/path/to/the/pdf/file(s')
    >>> obj.run()  # Start the analysis and extraction of data.

    Setting the parameter *auto* to true in the constructor will \
    automatically start the :meth:`run <.ElectoralRoll.run>` method:

    >>> obj = ElectoralRoll(source='/path/to/the/pdf/file(s', auto=True)

    See the :meth:`run <.ElectoralRoll.run>` at that method for a better \
    understanding of this class.
    '''

    inner_class_parser = RollParser
    inner_class_adapter = RollAdapter
    inner_class_printer = RollPrinter
    inner_class_memorizer = RollMemorizer
    inner_class_exporter = RollExporter

    # Operational methods
    # --------------------
    def run(self):
        '''
        :meth:`ElectoralRoll.run <.ElectoralRoll.run>` is the main method \
        within the class logic that executes the complete flow of data \
        analysis and extraction:

        * Iterate over the found files, ordered by size from smallest \
        to largest, executing the :meth:`run_file <.ElectoralRoll.run_file>` \
        method with the file, its index and the total.
        * It iterates on each page of each file:
            * *Processing* it with the library determined in the processor \
            property and defined in the constructor (see more in \
            :class:`PDFProcessorMixin <.PDFProcessorMixin>`).
            * *Adapting* the rendered page if required by the application \
            and the processor used (see more in :class:`RollAdapter \
            <.RollAdapter>`).
            * *Analyzing* the content text to extract its data (see more \
            in :class:`RollParser <.RollParser>`).
            * *Memorizing* your data in a consolidated data stored in the \
            memorizer. Its execution can be skipped by setting the \
            *memorizer* parameter to false in the constructor (see more \
            in :class:`RollMemorizer <.RollMemorizer>`).
            * *Exporting* your data to one or more csv files depending on \
            how the exporter is configured. Its execution can be activated \
            by defining the *export* parameter as true in the constructor \
            (see more in :class:`RollExporter <RollExporter>`).
        * The printer (:class:`RollPrinter <.RollPrinter>` instance) is \
        executed in each part of the flow and it determines if and how it \
        prints on the screen (as declared in the constructor).

        >>> roll.run()
        '''
        started = dt.now()
        self._metadata['analysis']['started'] = started
        files = self._metadata['files']
        self.printer.run_started(started, files)
        files = [x[1] for x in sorted(
            files.items(), key=lambda x: x[1]['bytes'])]
        for idx, file in enumerate(files):
            self.run_file(file, idx, len(files))
        finalized = dt.now()
        self._metadata['analysis']['finalized'] = finalized
        summary = self.exporter.export_summary(self.rid, self.metadata)
        if summary:
            self._metadata['exported_to'].append(summary)
        self._is_runned = True
        self.printer.run_finalized(finalized, self.metadata)

    def run_file(self, file, file_num, file_total):
        '''
        :param dict file: data of file
        :param int file_num: the number of file to analize.
        :param int file_total: the total of files to analize.
            this param and before is needed for printer.

        The :meth:`run_file <.ElectoralRoll.run_file>` method is called by \
        the :meth:`run <.ElectoralRoll.run>` method and iterates on each page \
        of each file:

            * *Processing* it with the library determined in the processor \
            property and defined in the constructor (see more in \
            :class:`PDFProcessorMixin <.PDFProcessorMixin>`).
            * *Adapting* the rendered page if required by the application \
            and the processor used (see more in :class:`RollAdapter \
            <.RollAdapter>`).
            * *Analyzing* the content text to extract its data (see more \
            in :class:`RollParser <.RollParser>`).
            * *Memorizing* your data in a consolidated data stored in the \
            memorizer. Its execution can be skipped by setting the \
            *memorizer* parameter to false in the constructor (see more \
            in :class:`RollMemorizer <.RollMemorizer>`).
            * *Exporting* your data to one or more csv files depending on \
            how the exporter is configured. Its execution can be activated \
            by defining the *export* parameter as true in the constructor \
            (see more in :class:`RollExporter <RollExporter>`).

        Stores metadatas of the extraction of each file.
        '''

        def get_progress(self, rid, files, sheets):
            metadata = self.metadata['rolls'].get(rid, {})
            if not metadata:
                metadata['entries'] = {'total': 0, 'errors': 0}
            return {
                'entries': metadata['entries']['total'],
                'errors': metadata['entries']['errors'],
                'files': files, 'sheets': sheets,
                'duration': dt.now() - self.metadata['analysis']['started'],
                }

        def duration_wrapper(self, file, stage, method, args):
            init = dt.now()
            method = getattr(self, method)
            result = method(*args)
            duration = dt.now() - init
            self._metadata['files'][file]['durations'][stage] += duration
            self._metadata['analysis']['durations'][stage] += duration
            return result

        def update_file_metadata(parsed, metadata):
            if not metadata:
                metadata['rid'] = parsed.metadata['rid']
                attributes = ['roll', 'year', 'region',
                              'province', 'commune']
                for attr in attributes:
                    metadata[attr] = parsed.header[attr]
                metadata['entries'] = {'total': 0, 'rescue': 0, 'errors': 0}
                declared = parsed.header.get('total_sheets', False)
                if declared:
                    metadata['entries']['declared'] = declared
            entries = parsed.metadata['entries']
            for meta in entries:
                metadata['entries'][meta] += entries[meta]
            return metadata

        # pre-processing
        init = dt.now()
        pdf = self.process_pdf(file['absolute'])
        total_sheets = len(pdf)  # number of pages.
        file_metadata = {}
        rid = None
        self.printer.run_file_start(file, file_num)
        for idx, sheet in enumerate(pdf):
            # printing
            progress = get_progress(
                self, rid, (file_num, file_total), (idx+1, total_sheets))
            self.printer.run_file_progress(progress)
            # processing
            processed = duration_wrapper(
                self, file['name'], 'processing', 'process_pdf_page', [sheet])
            # adapting
            adapted = duration_wrapper(
                self, file['name'], 'adapting', 'inner_class_adapter',
                [processed, self.processor]).sheet
            # parsingd
            parsed = duration_wrapper(
                self, file['name'], 'parsing', 'sheet_parse', [adapted])
            # memorizing
            duration_wrapper(self, file['name'], 'memorizing',
                             'sheet_memorize', [parsed])
            # exporting
            exported = duration_wrapper(
                self, file['name'], 'exporting', 'sheet_export', [parsed])
            if exported:
                if 'exported_to' not in self.metadata:
                    self._metadata['exported_to'] = []
                if exported not in self._metadata['exported_to']:
                    self._metadata['exported_to'].append(exported)
            # update file metadata
            file_metadata = update_file_metadata(parsed, file_metadata)
            rid = file_metadata['rid']
        file_metadata['duration'] = dt.now() - init
        self._metadata['files'][file['name']].update(file_metadata)
        self.printer.run_file_end(file_metadata)

    def sheet_parse(self, sheet, *args, **kwargs):
        '''
        :param str sheet: sheet in string.
        :return: instance of :class:`RollParser <.RollParser>`.

        Method that calls the class defined in the :attr:`inner_class_parser \
        <.ElectoralRoll.inner_class_parser>` class attribute, initializing it \
        with the sheet argument.
        '''
        return self.inner_class_parser(sheet, *args, **kwargs)

    def sheet_memorize(self, parsed, *args, **kwargs):
        '''
        :param str parsed: instance of :class:`RollParser <.RollParser>`.

        Method that routes a parsed page to the :meth:`memorize \
        <.RollMemorizer.memorize>` method of the memorizer.
        '''
        return self.memorizer.memorize(parsed, *args, **kwargs)

    def sheet_export(self, parsed, *args, **kwargs):
        '''
        :param str parsed: instance of :class:`RollParser <.RollParser>`.
        :return: Absolute path of the csv file where the data was exported.

        Method that routes a parsed page to the :meth:`export_sheet \
        <.RollExporter.export_sheet>` method of the exporter.
        '''
        return self.exporter.export_sheet(parsed, *args, **kwargs)

    @property
    def printer(self):
        '''
        :return: inner instance of :class:`RollPrinter <.RollPrinter>`.

        Property to call the :class:`RollPrinter <.RollPrinter>` object \
        instanciated in constructor.

        >>> roll.printer.__class__
        serveliza.roll.printer.RollPrinter
        '''
        return self._printer

    @property
    def memorizer(self):
        '''
        :return: inner instance of :class:`RollMemorizer <.RollMemorizer>`.

        Property to call the :class:`RollMemorizer <.RollMemorizer>` object \
        instanciated in constructor.

        >>> roll.memorizer.__class__
        serveliza.roll.memorizer.RollMemorizer
        '''
        return self._memorizer

    @property
    def exporter(self):
        '''
        :return: inner instance of :class:`RollExporter <.RollExporter>`.

        Property to call the :class:`RollExporter <.RollExporter>` object \
        instanciated in constructor.

        >>> roll.exporter.__class__
        serveliza.roll.exporter.RollExporter
        '''
        return self._exporter

    # Operational properties
    # -----------------------
    @property
    def is_runned(self):
        '''
        :return: boolean.

        Boolean property that indicates whether the instance has executed the \
        :meth:`run <.ElectoralRoll.run>` method or not.

        >>> roll.is_runned
        True  # or False
        '''
        return self._is_runned

    @property
    def metadata(self):
        '''
        :return: dictionary with all metadata.

        Property that stores the analysis metadata.
        It integrates the metadata of each electoral register detected \
        in the analysis.

        >>> roll.is_runned
        False
        >>> roll.metadata
        {'files': {'filename.pdf': {'name': 'filename.pdf',
           'bytes': 10000,
           'relative': 'relative/path/filename.pdf',
           'absolute': '/absolute/path/filename.pdf',
           'mtime': datetime.datetime(...),
           'atime': datetime.datetime(...),
           'durations': {'processing': datetime.timedelta(0),
            'adapting': datetime.timedelta(0),
            'parsing': datetime.timedelta(0),
            'memorizing': datetime.timedelta(0),
            'exporting': datetime.timedelta(0)}}},
         'analysis': {'started': None,
          'finalized': None,
          'durations': {'processing': datetime.timedelta(0),
           'adapting': datetime.timedelta(0),
           'parsing': datetime.timedelta(0),
           'memorizing': datetime.timedelta(0),
           'exporting': datetime.timedelta(0)}},
         'rolls': {}}
        >>> roll.run()
        >>> roll.metadata
        {'files': {'filename.pdf': {'name': 'filename.pdf',
           ...
           'rid': 'RID-XXXX',
           'roll': 'PADRON ELECTORAL X - ELECCIONES X XXXX',
           'year': XXXX,
           'region': 'REGION',
           'province': 'PROVINCE',
           'commune': 'COMMUNE',
           'entries': {'total': 999, 'rescue': 0, 'errors': 0},
           'duration': datetime.timedelta(...)}},
         'analysis': {'started': datetime.datetime(...) ...},
         'rolls': {'RID-XXXX: {'roll': 'PADRON ELECTORAL X - ELE...',
           'year': XXXX,
           'regions': ['REGION', ...],
           'communes': ['COMMUNE', ...],
           'provinces': ['PROVINCE', ...],
           'nulls': {'total': 0},
           'entries': {'total': 999, 'rescue': 0, 'errors': 0}}}}
        '''
        stored = {'rolls': {}}
        for k, v in self.memorizer.storage.items():
            if 'metadata' in v:
                stored['rolls'][k] = v['metadata']
        return {**self._metadata, **stored}

    @property
    def rid(self):
        '''
        :return: string with first roll identifier.

        Property that returns the identifier of the electoral roll analyzed.
        If it will return only the first identifier detected, this should \
        not cause inconvenience unless pdf files from different electoral \
        rolls are loaded.

        >>> roll.rid
        'RID-XXXX'

        If the instance did not run, it returns None.
        '''
        if self.memorizer.storage:
            rids = [x for x in self.memorizer.storage]
            return rids[0]

    @property
    def roll(self):
        '''
        :return: name of electoral roll.

        Property that returns the full name of electoral roll analyzed.

        >>> roll.roll
        'PADRON ELECTORAL X - ELECCIONES X XXXX'

        Internaly use the :attr:`rid <.ElectoralRoll.rid>` property. If the \
        instance did not run, it returns None.
        '''
        if self.memorizer.storage:
            return self.memorizer.storage[self.rid]['metadata']['roll']

    @property
    def entries(self):
        '''
        :return: list of data entries of memorizer.

        Property that accesses the data entries of the electoral roll \
        analyzed. The data is stored in the :class:`RollMemorizer \
        <.RollMemorizer>` instance.

        >>> roll.entries
        [[...]...]

        Internaly use the :attr:`rid <.ElectoralRoll.rid>` property. If the \
        instance did not run, it returns None.
        '''
        storage = self.memorizer.storage
        if storage and 'entries' in storage[self.rid]:
            return self.memorizer.storage[self.rid]['entries']

    @property
    def fields(self):
        '''
        :return: list of fields of electoral roll.

        Property that returns the fields of the electoral roll analyzed.

        >>> roll.fields
        ['nombre',
         'c-identidad',
         'sex',
         'region',
         'provincia',
         'comuna',
         'domicilio-electoral',
         'circunscripcion',
         'mesa',
         'reference']

        Internaly use the :attr:`rid <.ElectoralRoll.rid>` property. If the \
        instance did not run, it returns None.
        '''
        storage = self.memorizer.storage
        if storage and 'fields' in storage[self.rid]:
            return self.memorizer.storage[self.rid]['fields']

    @property
    def errors(self):
        '''
        :return: list of errors found.

        Property that stores the errors of the analysis. List of errors \
        found in the analysis. Errors are dictionaries with data to keep \
        track of. The purpose of registering them is to improve the \
        development of serveliza.

        >>> roll.errors
        [...]
        '''
        return self.memorizer.errors

    @property
    def to_dataframe(self):
        '''
        :return: Pandas DataFrame instance.
        :raises UserWarning: You need to run the application before \
            converting the result to Pandas DataFrame.

        Property that returns the electoral roll data in a new Pandas \
        `DataFrame`_ instance.

        .. _DataFrame: https://pandas.pydata.org/pandas-docs/stable/\
            reference/api/pandas.DataFrame.html
        '''
        if not self.is_runned:
            raise UserWarning('You need to run the application before '
                              'converting the result to Pandas DataFrame.')
        return pd.DataFrame(np.array(self.entries), columns=self.fields)

    @property
    def source(self):
        '''
        :return: list of paths to valid pdf files.
        :raises TypeError: source param must be string or list.
        :raises TypeError: source doesnt have valid PDF files.

        Property that stores paths of pdf files obtained from a list or \
        string with file paths or directories.

        >>> roll.source
        ['relative / path / to / file.pdf']

        The source is loaded into the constructor through the parameter \
        of the same name. It is also possible to redefine through the \
        property setter:

        >>> roll.source = ['path / to / file.pdf', '/ path / to / dir']
        >>> roll.source = '/path/to/dir/o/file.pdf'
        '''
        return self._source

    @source.setter
    def source(self, source):
        paths = []
        if isinstance(source, str):
            paths.append(source)
        elif isinstance(source, list):
            paths += source
        else:
            raise TypeError('source param must be string or list.')
        files = []
        for path in paths:
            if isinstance(path, str) and pdf_utils.is_valid_pdf(path):
                files += [path]
            elif isinstance(path, str) and Path(path).is_dir():
                files += self.printer.init_search(
                    pdf_utils.get_all_pdf_in_path, [path, self.recursive])
        if not files:
            raise TypeError('Source doesnt have valid PDF files.')
        self._source += files
        meta_files = pdf_utils.get_metadata_from_pdfs(files)
        self.printer.init_founded(meta_files)
        for file in meta_files:
            meta_files[file]['durations'] = DURATIONS_SCHEMA
        self._metadata['files'].update(meta_files)

    @property
    def recursive(self):
        '''
        Property that determines if the search for pdf files in the \
        delivered source is recursive or is only for the root of the \
        indicated directory,
        '''
        return self._recursive

    def __init__(self, source, auto=False, *args, **kwargs):
        processor = kwargs.get('processor', self.processor)
        self.processor = processor
        self._printer = self.inner_class_printer(**kwargs)
        self._memorizer = self.inner_class_memorizer(**kwargs)
        self._exporter = self.inner_class_exporter(**kwargs)
        self._metadata = {'files': {}}
        self._is_runned = False
        self._recursive = bool(kwargs.get('recursive', False))
        self._source = []
        self.source = source
        self._metadata['analysis'] = {
            'started': None,
            'finalized': None,
            'durations': DURATIONS_SCHEMA}
        if auto:
            self.printer.init_auto()
            self.run()

    def __repr__(self):
        return self.printer.repr(self)




from pdfminer.layout import LTTextBoxHorizontal


class RollNoisedError(Exception):
    '''
    Exception that indicates an error when analyzing because the \
    pattern is noisy, that is, it has watermarks to prevent the \
    coherent extraction of text.
    '''
    pass


class PdftotextAdapterMixin:
    '''
    :class:`PdftotextAdapterMixin <.PdftotextAdapterMixin>` is an adapter for \
    the `pdftotext <https://github.com/jalan/pdftotext>`_ processor.

    It is a mixin designed to be inherited in :class:`RollAdapter \
    <.RollAdapter>`.
    '''

    def adapter_pdftotext(self, sheet):
        '''
        :param str sheet: sheet in text string.
        :raises ValueError: Unexpected type of sheet.
        :raises RollNoisedError: pdftotext processor cant process a \
            noised roll. Try with the pdfminersix processor.
        :return: sheet adapted.

        Method to adapt a sheet processed by `pdftotext \
        <https://github.com/jalan/pdftotext>`_ before being \
        passed to the parser.
        '''
        if not isinstance(sheet, str):
            raise ValueError('Unexpected type of sheet.')
        if len(sheet) > 100000:
            # indicate the pdf file is noised.
            raise RollNoisedError(
                'pdftotext processor cant process a noised roll. '
                'Try with the pdfminersix processor.')
        return sheet


class PdfminersixAdapterMixin:
    '''
    :class:`PdfminersixAdapterMixin <.PdfminersixAdapterMixin>` is an adapter \
    for the `pdfminersix <https://pdfminersix.readthedocs.io/>`_ processor.

    It is a mixin designed to be inherited in :class:`RollAdapter \
    <.RollAdapter>`.
    '''

    def adapter_pdfminersix(self, sheet):
        '''
        :param list sheet: sheet in list of pdfminersix elements.
        :return: sheet adapted in text string.

        Method to adapt a sheet processed by `pdfminersix \
        <https://pdfminersix.readthedocs.io/>`_ before being \
        passed to the parser.

        It is also capable of eliminating possible noise with watermarks.
        '''
        # purge the noise
        filtered = [x for x in sheet if isinstance(
            x, LTTextBoxHorizontal) and x.bbox[0] > 0]
        # group header & columns
        layout = {}
        elements = []
        for element in filtered:
            elements += [x for x in element]
        for element in elements:
            if element.bbox[1] not in layout:
                layout[element.bbox[1]] = [element]
            else:
                layout[element.bbox[1]].append(element)
        # ordering and serialization
        for height in layout:
            layout[height].sort(key=lambda x: x.bbox[0])
            stringed = ' \t '.join([x.get_text(
                ).replace('\n', '') for x in layout[height]]) + '\n'
            layout[height] = stringed
        # return in one string
        return ''.join([x[1] for x in sorted(
            layout.items(), key=lambda x:x[0], reverse=True)])


ADAPTERS = [PdftotextAdapterMixin, PdfminersixAdapterMixin]


class RollAdapter(*ADAPTERS):
    '''
    :param obj sheet: sheet of the type according to the processor used.
    :param str processor: processor used in the sheet.

    :class:`RollAdapter <.RollAdapter>` is a class that is instantiated \
    for a sheet by routing it to the appropriate method according to \
    the processor defined in the constructor parameter of the same name.

    Practical use:

    >>> adapted = RollAdapted(processed_sheet, 'processor-name').sheet
    '''
    @property
    def sheet(self):
        '''
        Property where the adapted sheet is stored in the constructor.
        '''
        return self._sheet

    def __init__(self, sheet, processor, *args, **kwargs):
        if not isinstance(processor, str):
            raise ValueError('processor must be string.')
        adapter = getattr(self, 'adapter_'+processor, None)
        if not adapter:
            raise ValueError(f'Processor {processor} dont have valid method.')
        self._sheet = adapter(sheet)




class RollMemorizer:
    '''
    :param bool memorize: If the memorizer is activated (default True)

    :class:`RollMemorizer <.RollMemorizer>` is a class that allows it \
    to store data and errors from the electoral roll. It is instantiated \
    within an instance of :class:`ElectoralRoll <.ElectoralRoll>`.
    '''

    @property
    def storage(self):
        '''
        :return: dictionary with all data.

        Property where all the memorized data are stored.
        '''
        return self._storage

    @property
    def errors(self):
        '''
        :return: list with errors.

        Property where the errors found are stored.
        '''
        return self._errors

    @property
    def is_active(self):
        '''
        :return: boolean.

        Property that indicates if the memorizer is active as defined \
        in the constructor.
        '''
        return self._is_active

    def memorize(self, parsed):
        '''
        :param obj parsed: an instance of :class:`RollParser <.RollParser>`.

        :meth:`memorize <.RollMemorizer.memorize>` is the main method of \
        :class:`RollMemorizer <.RollMemorizer>`. It will memorize the \
        metadata of an analyzed sheet, if the memorizer is active (see \
        :attr:`is_active <.RollMemorizer.is_active>` property) it will \
        also memorize the analyzed entries. The methods executed to \
        memorize the metadata in order:

        * :meth:`prepare_rid \
            <.RollMemorizer.prepare_rid>`.
        * :meth:`store_metadata_places \
            <.RollMemorizer.store_metadata_places>`.
        * :meth:`store_metadata_entries \
            <.RollMemorizer.store_metadata_entries>`.
        * :meth:`store_metadata_nulls \
            <.RollMemorizer.store_metadata_nulls>`.
        * :meth:`store_metadata_nulls \
            <.RollMemorizer.store_metadata_nulls>`.

        It then stores, if active, the entries and errors.
        '''
        rid = parsed.metadata['rid']
        self.prepare_rid(parsed)
        self.store_metadata_places(parsed)
        self.store_metadata_entries(parsed)
        self.store_metadata_nulls(parsed)
        if self.is_active:
            self._storage[rid]['entries'] += parsed.entries
            self._errors += parsed.errors

    def prepare_rid(self, parsed):
        '''
        :param obj parsed: an instance of :class:`RollParser <.RollParser>`.

        :meth:`prepare_rid <.RollMemorizer.prepare_rid>` is a method that \
        prepares the :attr:`storage <.RollMemorizer.storage>` property for \
        storing metadata and electoral roll data. Use the electoral *roll \
        identifier* (see more in :attr:`rid <.ElectoralRoll.rid>`) as the \
        key for the :attr:`storage <.RollMemorizer.storage>` property.
        '''
        rid = parsed.metadata['rid']
        if rid in self.storage:
            return None
        metadata = {
            'roll': parsed.header['roll'],
            'year':     parsed.header['year'],
            'regions':      [],
            'communes':     [],
            'provinces':    [],
            'nulls': {'total': 0},
            'entries': {'total': 0, 'rescue': 0, 'errors': 0}}
        declared = parsed.header.get('total_sheets', False)
        if declared:
            metadata['entries']['declared'] = declared
        self._storage[rid] = {
            'entries':  [],
            'fields':   parsed.fields,
            'metadata': metadata}

    def store_metadata_places(self, parsed):
        '''
        :param obj parsed: an instance of :class:`RollParser <.RollParser>`.

        :meth:`store_metadata_places <.RollMemorizer.store_metadata_places>` \
        is a method to memorize the places (regions, provinces and communes) \
        present in the parsed sheet.
        '''
        rid = parsed.metadata['rid']
        region = parsed.header['region']
        regions = self.storage[rid]['metadata']['regions']
        province = parsed.header['province']
        provinces = self.storage[rid]['metadata']['provinces']
        commune = parsed.header['commune']
        communes = self.storage[rid]['metadata']['communes']
        if region not in regions:
            self._storage[rid]['metadata']['regions']\
                .append(region)
        if province not in provinces:
            self._storage[rid]['metadata']['provinces']\
                .append(province)
        if commune not in communes:
            self._storage[rid]['metadata']['communes']\
                .append(commune)

    def store_metadata_entries(self, parsed):
        '''
        :param obj parsed: an instance of :class:`RollParser <.RollParser>`.

        :meth:`store_metadata_entries <.RollMemorizer.store_metadata_entries>`\
         is a method to memorize the metadata of the entries (*total*, \
         *rescued*, *errors*) present in the parsed sheet. If the total \
         number of entries is declared in the header, it is added as \
         *declared*.
        '''
        rid = parsed.metadata['rid']
        entries = self.storage[rid]['metadata']['entries']
        pentries = parsed.metadata['entries']
        total = entries['total'] + len(parsed.entries)
        rescue = entries['rescue'] + pentries.get('rescue', 0)
        errors = entries['errors'] + len(parsed.errors)
        final_entries = {'total': total, 'rescue': rescue, 'errors': errors}
        declared = entries.get('declared', False)
        if declared:
            final_entries['declared'] = declared
        self._storage[rid]['metadata']['entries'] = final_entries

    def store_metadata_nulls(self, parsed):
        '''
        :param obj parsed: an instance of :class:`RollParser <.RollParser>`.

        :meth:`store_metadata_nulls <.RollMemorizer.store_metadata_nulls>` \
        is a method to memorize the metadata of the null data in the entries \
        (total and for each field with null data) present in the parsed sheet.
        '''
        rid = parsed.metadata['rid']
        pnulls = parsed.metadata['nulls']
        if not pnulls['total']:
            return None
        nulls = self.storage[rid]['metadata']['nulls']
        total = nulls['total'] + pnulls['total']
        fields = parsed.metadata['fields']
        result = {'total': total}
        for field in fields:
            if field not in pnulls:
                continue
            result[field] = pnulls[field]
            if field in nulls:
                result[field] += nulls[field]
        self._storage[rid]['metadata']['nulls'].update(result)

    def __init__(self, *args, **kwargs):
        memorize = kwargs.get('memorize', True)
        self._is_active = bool(memorize)
        self._storage, self._errors = {}, []



import os
from serveliza.utils import humanize


class ColorMixin:
    '''
    Mixin that grants methods to color text if the colors property is true.
    '''
    def info(self, text):
        '''
        Color an information text (blue).
        '''
        if self.colors:
            return f'\033[0;94m{str(text)}\033[0m'
        return text

    def ok(self, text):
        '''
        Color an success text (green).
        '''
        if self.colors:
            return f'\033[0;32m{str(text)}\033[0m'
        return text

    def warn(self, text):
        '''
        Color an warn text (yellow).
        '''
        if self.colors:
            return f'\033[0;33m{str(text)}\033[0m'
        return text

    def subtle(self, text):
        '''
        Color an sublte text (gray).
        '''
        if self.colors:
            return f'\033[0;90m{str(text)}\033[0m'
        return text

    def error(self, text):
        '''
        Color an error text (red).
        '''
        if self.colors:
            return f'\033[0;31m{str(text)}\033[0m'
        return text


class RollPrinter(ColorMixin):
    '''
    :param bool verbose: If print the progress in screen (default False).
    :param bool colors: If print with colors in the screen (default True).

    :class:`RollPrinter <.RollPrinter>` is a class that allows it \
    to print progress of application in the screen. It is instantiated \
    within an instance of :class:`ElectoralRoll <.ElectoralRoll>`.
    '''

    @property
    def verbose(self):
        '''
        Property that determines whether to print the application \
        progress to the screen.
        '''
        return self._verbose

    @property
    def colors(self):
        '''
        Property that determines whether to print on screen with colors.
        '''
        return self._colors

    def init_search(self, func, args):
        '''
        Method that prints on constructor search.
        '''
        if not self.verbose:
            return func(*args)
        msg = self.info('Searching')
        if args[1]:
            msg += self.subtle(' (recursively) ')
        msg += self.subtle('... ')
        print(msg, end='')
        result = func(*args)
        if result:
            print(self.ok('OK'))
        return result

    def init_founded(self, files):
        '''
        Method that prints the search result of the constructor.
        '''
        if not self.verbose:
            return None
        plural = 's' if len(files) > 1 else ''
        msg = self.info(f'Founded {len(files)} pdf file{plural}: \n')
        for file, meta in files.items():
            size = humanize.this_bytes(meta['bytes'])
            msg += ' '+self.ok(file)+' - '
            msg += self.subtle(f'({size}) {meta["relative"]}')+'\n'
        print(msg)

    def init_auto(self):
        '''
        Method that prints if the start was automatic.
        '''
        if not self.verbose:
            return None
        print(self.warn('Running automatically'))

    def run_started(self, started, files):
        '''
        Method that prints if the start of the analysis.
        '''
        if not self.verbose:
            return None
        msg = self.info(f'Running analysis of electoral roll'
                        f' in {str(len(files))} pdf files')+'\n'
        msg += self.subtle(f'At {str(started)[:-7]}.')+'\n'
        msg += self.info('Each file will be processed, adapted, '
                         'analyzed, memorized and exported in a '
                         'single cycle for better performance. '
                         'It will start with the smallest files.')+'\n'
        print(msg)

    def run_file_start(self, file, number):
        '''
        Method that prints if the start of the analysis of a file.
        '''
        if not self.verbose:
            return None
        size = humanize.this_bytes(file['bytes'])
        msg = f'\r> {str(number)}: {self.ok(file["name"])} '
        msg += self.subtle(f' ({size}) ')
        msg += self.subtle(f' {file["relative"]}')
        self.clean_line()
        print(msg)

    def run_file_progress(self, pro):
        '''
        Method that prints the progress of the analysis of a file.
        '''
        if not self.verbose:
            return None
        # spinner cycle
        if self._tmp_run_count > 3:
            self._tmp_run_count = 0
        cycle = self.info(r'-\|/'[self._tmp_run_count])
        self._tmp_run_count += 1
        entries = self.ok(
            f"{pro['entries']}") if pro['entries'] else '0'
        errors = self.error(
            f"{pro['errors']}") if pro['errors'] else '0'
        self.clean_line()
        print(f'\r{cycle} Files: {pro["files"][0]}/{pro["files"][1]} '
              f'[{self.percent(*pro["files"])}%]',
              f'Sheet: {str(pro["sheets"][0])}/{str(pro["sheets"][1])} '
              f'[{self.percent(*pro["sheets"])}%] ',
              f'[{entries}/{errors}]',
              f'Time: {str(pro["duration"])[:-7]}',
              end='', sep=' | ', flush=True)

    def run_file_end(self, metadata):
        '''
        Method that prints the completion of the analysis of a file.
        '''
        if not self.verbose:
            return None
        self.clean_line()
        msg = self.subtle('\r- ') + self.info(metadata['rid'])
        roll = self.subtle(metadata['roll'][:17]+'...'+metadata['roll'][-10:])
        commune = metadata['commune']
        meta_entries = metadata['entries']
        entries = self.ok(
            meta_entries['total']) if meta_entries['total'] else '0'
        errors = self.error(
            meta_entries['errors']) if meta_entries['errors'] else '0'
        msg += f' {roll} > {self.info(commune)} '
        msg += f'({entries}/{errors})'
        msg += f' | {str(metadata["duration"])[:-7]}'
        print(msg)

    def run_finalized(self, finalized, metadata):
        '''
        Method that prints the completion of the analysis.
        '''
        if not self.verbose:
            return None
        files_num = len(metadata['files'])
        analysis = metadata['analysis']
        duration = analysis['finalized'] - analysis['started']
        durations = analysis['durations']
        rolls = metadata['rolls']
        msg = ''
        for rid, roll in rolls.items():
            re = [x for x in roll['regions'] if x]
            pro = [x for x in roll['provinces'] if x]
            com = [x for x in roll['communes'] if x]
            ent = roll['entries']
            regions = ','.join(re) if len(re) < 3 else str(len(re))
            provinces = ','.join(pro) if len(pro) < 3 else str(len(pro))
            communes = ','.join(com) if len(com) < 3 else str(len(com))
            msg += self.info('Roll: ')+f'{roll["roll"]}.\n'
            msg += self.info('Entries: ')+self.ok(
                f"{ent['total']:_}".replace('_', '.'))+'\t'
            msg += self.info('Errors: ')+self.error(ent['errors'])+'\n'
            msg += self.info('Region(s): ')+regions+'.\n'
            msg += self.info('Province(s): ')+provinces+'.\n'
            msg += self.info('Commune(s): ')+communes+'.\n'
        if 'exported_to' in metadata:
            msg += '\nExported to:'
            for path in metadata['exported_to']:
                msg += self.ok(f'\n > {path}')
        msg += self.info(
            f'\n\nAnalysis of {str(files_num)} files finished.')+'\n'
        msg += self.subtle(f'At {str(finalized)[:-7]}.')
        msg += '\n'+self.warn('Duration: ')+f'{str(duration)[:-7]}'
        msg += self.subtle(
            f'\n- processing: {str(durations["processing"])}'
            f'\n- adapting: {str(durations["adapting"])}'
            f'\n- parsing: {str(durations["parsing"])}')
        if durations['memorizing']:
            msg += self.subtle(
                f'\n- memorizing: {str(durations["memorizing"])}')
        if durations['exporting']:
            msg += self.subtle(
                f'\n- exporting: {str(durations["exporting"])}')
        print(msg)

    def percent(self, of, total):
        '''
        Utility that returns a percentage in string. It receives \
        two parameters (of and total) with which the relation calculates.
        '''
        return str(int((100/total)*of))

    def clean_line(self):
        '''
        Utility that cleans the last line printed on the screen.
        '''
        print('\r'+' '*os.get_terminal_size().columns, end='')

    def repr(self, obj):
        '''
        Method to print representation of :class:`ElectoralRoll \
        <.ElectoralRoll>` class.
        '''
        msg = '<' + self.info(obj.__class__.__name__)
        msg += ' instance ' + self.is_runned_tag(obj.is_runned)
        if obj.is_runned:
            entries = self.ok(
                len(obj.entries)) if len(obj.entries) else '0'
            errors = self.error(
                len(obj.errors)) if len(obj.errors) else '0'
            msg += f'[{self.ok(obj.rid)}]'
            msg += f'({entries}/{errors})'
            msg += f'[{(len(obj.metadata["files"]))} files]'
        msg += '>'
        return msg

    def is_runned_tag(self, is_runned):
        '''
        Utility returns a text string formatted to print if \
        :class:`ElectoralRol <.ElectoralRoll>` ran.
        '''
        if is_runned:
            return f"[{self.ok('is_runned')}]"
        return f"[{self.warn('not is_runned')}]"

    def __init__(self, *args, **kwargs):
        # temp attrs
        self._tmp_run_count = 0
        self._tmp_run_file_progress_dt = None
        if 'verbose' in kwargs:
            self._verbose = bool(kwargs['verbose'])
        else:
            self._verbose = False
        if 'colors' in kwargs:
            self._colors = bool(kwargs['colors'])
        self._colors = True
        self._log = []




def this_bytes(bts, short=False):
    if not isinstance(bts, int):
        return None
    measures = [
        (1000000000, 'gigabytes', 'gb'),
        (1000000, 'megabytes', 'mb'),
        (1000, 'kilobytes', 'kb'),
        (1, 'bytes', 'b')
    ]
    for measure in sorted(measures, reverse=True, key=lambda x: x[0]):
        if bts > measure[0]:
            msr = measure[1] if not short else measure[2]
            return str(bts // measure[0]) + ' ' + msr



# PDF processors libraries:
# --------------------------
import pdftotext
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.layout import LAParams
from pdfminer.converter import PDFPageAggregator


class PdftotextMixin:
    '''
    '''
    def processor_pdftotext(self, pathfile):
        '''
        Method to use `pdftotext <https://github.com/jalan/pdftotext>`_ \
        in a file specified in the argument as a path.

        >>> obj.processor_pdftotext('/path/to/file.pdf')
        list # without processing
        '''
        self._tmp_file = open(str(pathfile), 'rb')
        return pdftotext.PDF(self._tmp_file)

    def processor_pdftotext_page(self, page):
        '''pdftotext not need that.'''
        return page


class PdfminersixMixin:
    '''
    '''
    def processor_pdfminersix(self, pathfile):
        '''
        '''
        self._tmp_file = open(str(pathfile), 'rb')
        return [x for x in PDFPage.get_pages(self._tmp_file)]

    def processor_pdfminersix_page(self, page):
        '''
        '''
        resource_manager = PDFResourceManager()
        device = PDFPageAggregator(resource_manager, laparams=LAParams())
        interpreter = PDFPageInterpreter(resource_manager, device)
        interpreter.process_page(page)
        return device.get_result()


PROCESSORS = [PdftotextMixin, PdfminersixMixin]




from urllib.error import HTTPError
import pandas as pd
import json

CIR_PRES_89_17_NAC_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_presidenciales_ce_1989_2017_Chile.xlsx'
CIR_PRES_89_17_EXT_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_presidenciales_ce_2017_Extranjero.xlsx'
CIR_SENA_89_17_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_senadores_ce_1989_2017.xlsx'
CIR_DIPU_89_17_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_diputados_ce_1989_2017.xlsx'
CIR_ALCA_04_16_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_alcaldes_ce_2004_2016.xlsx'
CIR_CONC_04_16_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_concejales_ce_2004_2016.xlsx'
CIR_ALCO_92_00_URL = 'https://www.servel.cl/wp-content/uploads/' \
    '2019/04/resultados_elecciones_concejales_ce_1992_al_2000.xlsx'

CIR_ALL_URLS = [
    CIR_PRES_89_17_NAC_URL,
    # CIR_PRES_89_17_EXT_URL, don't have a commune column
    CIR_SENA_89_17_URL,
    CIR_DIPU_89_17_URL,
    CIR_ALCA_04_16_URL,
    CIR_CONC_04_16_URL,
    CIR_ALCO_92_00_URL,
]

KEY_YEAR = 'Año de Elección'
KEY_REGION = 'Región'
KEY_PROVINCE = 'Nombre Provincia'
KEY_COMMUNE = 'Comuna'
KEY_CIR = 'Circunscripción Electoral'
KEY_CIR2 = 'Nombre Circunscripción Electoral'


def get_communes_circuns(url=None, verbose=False, export_json=False):
    '''
    Function to get from the Internet (servel.cl) a \
    dictionary object with comunne in key and list of \
    circunscriptions in value.

    If dont specify the url then take all urls in \
    CIR_ALL_URLS to get the data and consolidate \
    in unique result.

    If verbose is true then print the process in \
    console step by step.

    This function is used to create json fixtures \
    of communes and circuns with the export_json \
    param to true.
    '''
    def __url_to_dic(url):
        if verbose:
            print(f'''Waiting de XLS file from: {str(url)}''')
        try:
            doc = pd.read_excel(url)
        except HTTPError as e:
            print(f'''HTTP Error: {str(e.code)} -> {e.msg}''')
            print(f'''url: {url}''')
            return None
        if verbose:
            print('''xls file is downloaded''')
        if KEY_CIR2 in doc:
            key_cir = KEY_CIR2
        else:
            key_cir = KEY_CIR
        data = doc[[KEY_COMMUNE, key_cir]].drop_duplicates()
        data = data.to_dict('index').items()
        return [{item[KEY_COMMUNE]:item[key_cir]} for key, item in data]
    main_data = []
    if not url:
        if verbose:
            print('''Running in all urls availables''')
        for i, u in enumerate(CIR_ALL_URLS):
            if verbose:
                print(
                    'Getting the xls file '
                    f'{str(i+1)}/{str(len(CIR_ALL_URLS))}')
            dic = __url_to_dic(u)
            if not dic:
                if verbose:
                    print('''Not data found''')
                continue
            if verbose:
                print(f'Partial data founded {str(len(dic))} circunscriptions')
            main_data += dic
    else:
        if verbose:
            print(f'''Running in url: {str(url)}''')
        main_data += __url_to_dic(url)
    if not main_data:
        print('''Data from communes and circunscription are not availables''')
        return None
    if verbose:
        print(f'Data founded: {str(len(main_data))} circunscriptions')
    result = {}
    for item in main_data:
        for k, v in item.items():
            if k in result:
                if isinstance(v, float):
                    continue
                if v not in result[k]:
                    result[k].append(v)
            elif not isinstance(v, float):
                result[k] = [v]
    if verbose:
        print(f'Total communes: {str(len(result))}')
        print(
            'Total circunscriptions: '
            f'{sum([len(x) for x in result.values()])}')
    if export_json:
        return json.dumps(result)
    return result



# builtin libraries
import sys

# internal modules
from .pdf_processors import PROCESSORS

# from pdfminer.high_level import extract_pages

# disables warinings of processors libs
if not sys.warnoptions:
    import warnings
    warnings.simplefilter("ignore")


class PDFProcessorMixin(*PROCESSORS):
    '''
    Mixin that allows an instance the ability to process \
    PDF files with certain libraries. In the constructor, \
    the processor to be used is defined with the argument \
    of the same name, binding the :attr:`process_pdf \
    <.ProcesosrPDFMixin.process_pdf>` property to the \
    method related to it.

    PDF processor availables:

     - `pdftotext <https://github.com/jalan/pdftotext>`_ \
        (0.1.0 release) with :meth:`pdftotext_processor \
        <.ProcessorPDFMixin.processor_pdftotext>`
     - `pdfminersix <https://pdfminersix.readthedocs.io/>`_ \
        (0.1.0 release) with :meth:`pdftotext_processor \
        <.ProcessorPDFMixin.processor_pdfminersix>`
    '''
    _processor = 'pdftotext'
    _tmp_file = None
    processor_ref = {
        'pdftotext':   'https://github.com/jalan/pdftotext',  # 0.1.0
        'pdfminersix': 'https://pdfminersix.readthedocs.io/',  # 0.1.0
        # dev note: add processors ref here {name:url}.
        }

    @property
    def processor(self):
        '''
        Processor (library) to extract text from pdf file.
        '''
        return self._processor

    @processor.setter
    def processor(self, value):
        if value in self.processor_ref:
            self._processor = value
            self._process_pdf = getattr(self, 'processor_'+value)
            self._process_pdf_page = getattr(self, 'processor_'+value+'_page')
            self._processor_name = value
            return None
        raise TypeError(
            f'{str(value)} must be a available '
            'processor: ' + ','.join(
                [x[0] for x in self.processors_ref.items()]))

    @property
    def process_pdf(self):
        '''
        Property that calls the method corresponding to the PDF file \
        processor configured in the instance initialization.

        >>> obj.process_pdf(*args)
        '''
        if self._tmp_file:
            self._tmp_file.close()  # ensure closing file.
            self._tmp_file = None
        return self._process_pdf

    @property
    def process_pdf_page(self):
        '''
        '''
        return self._process_pdf_page



#!/usr/bin/env python

"""Tests for `serveliza` package."""

from datetime import datetime, timedelta
from pandas import pandas as pd
import unittest

from serveliza.roll import ElectoralRoll
from serveliza.roll.printer import RollPrinter
from serveliza.roll.memorizer import RollMemorizer
from serveliza.roll.exporter import RollExporter


class TestServeliza(unittest.TestCase):
    """Tests for `serveliza` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        pass

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_roll_2013(self):
        """Test something."""
        roll = ElectoralRoll(source='tests/fixtures/A0152003.pdf')
        self.roll_assert_props(roll)
        roll.run()
        self.roll_assert_runned(roll)

    def test_roll_2016(self):
        roll = ElectoralRoll(source='tests/fixtures/Antártica.pdf')
        self.roll_assert_props(roll)
        roll.run()
        self.roll_assert_runned(roll)

    def test_roll_2020(self):
        roll = ElectoralRoll(
            source='tests/fixtures/A12202.pdf',
            processor='pdfminersix')
        self.roll_assert_props(roll)
        roll.run()
        self.roll_assert_runned(roll)

    def roll_assert_props(self, roll):
        # operationals
        self.assertFalse(roll.is_runned)
        self.assertEqual(len(roll.metadata), 3)
        self.assertTrue('analysis' in roll.metadata)
        self.assertEqual(len(roll.metadata['analysis']), 3)
        self.assertTrue('started' in roll.metadata['analysis'])
        self.assertTrue('finalized' in roll.metadata['analysis'])
        self.assertEqual(len(roll.metadata['analysis']['durations']), 5)
        for stage, duration in roll.metadata['analysis']['durations'].items():
            self.assertTrue(isinstance(duration, timedelta))
        self.assertTrue('files' in roll.metadata)
        for file, meta in roll.metadata['files'].items():
            self.assertTrue('name' in meta)
            self.assertTrue('relative' in meta)
            self.assertTrue('absolute' in meta)
            self.assertTrue('mtime' in meta)
            self.assertTrue('atime' in meta)
            self.assertTrue('durations' in meta)
        self.assertTrue('rolls' in roll.metadata)
        self.assertTrue(isinstance(roll.printer, RollPrinter))
        self.assertTrue(isinstance(roll.memorizer, RollMemorizer))
        self.assertTrue(isinstance(roll.exporter, RollExporter))

    def roll_assert_runned(self, roll):
        self.assertTrue(roll.is_runned)
        # storage property
        self.assertTrue(bool(roll.rid))
        self.assertTrue(bool(roll.roll))
        self.assertTrue(bool(roll.entries))
        self.assertTrue(bool(roll.fields))
        self.assertTrue(isinstance(roll.to_dataframe, pd.DataFrame))
        metadata = roll.metadata['rolls'][roll.rid]
        self.assertTrue(len(metadata['regions']) > 0)
        self.assertTrue(len(metadata['communes']) > 0)
        self.assertTrue(len(metadata['provinces']) > 0)
        self.assertEqual(metadata['entries']['total'], len(roll.entries))
        # metadata property
        analysis = roll.metadata['analysis']
        self.assertTrue(isinstance(analysis['started'], datetime))
        self.assertTrue(isinstance(analysis['finalized'], datetime))
        self.assertTrue(analysis['durations']['processing'] > timedelta())
        self.assertTrue(analysis['durations']['adapting'] > timedelta())
        self.assertTrue(analysis['durations']['parsing'] > timedelta())
        self.assertTrue(analysis['durations']['memorizing'] > timedelta())
        self.assertTrue(analysis['durations']['exporting'] > timedelta())



"""Main module."""
from serveliza.roll import ElectoralRoll


def roll_from_pdf_to_csv(
        source, output='output',
        processor=None, mode=None, mode_sep=None,
        no_suffix=False, recursive=False, no_summary=False,
        silent=False, no_colors=False,):
    roll = ElectoralRoll(
        source=source, output=output,
        mode=mode, mode_sep=mode_sep,
        random_suffix=False if no_suffix else True,
        summary=False if no_summary else True,
        memorize=False, processor=processor,
        recursive=recursive,
        verbose=False if silent else True,
        colors=False if no_colors else True,
        export=True)
    roll.run()
    return roll.metadata['exported_to']


def roll_from_pdf_to_dataframe(
        source, recursive=False,
        verbose=False, processor=None):
    roll = ElectoralRoll(
        source=source, recursive=recursive,
        verbose=verbose, processor=None)
    roll.run()
    return roll.to_dataframe()



# builtin libraries
import os
import re
import json
from datetime import datetime as dt

# third party libraries
from slugify import slugify


class RollParser:
    '''
    :class:`RollParser <.RollParser>` is intended to be instantiated \
    by each sheet.

    Class attributes beginning with "*regex_*" correspond to the regular \
    expressions used to detect fields in the header. The *regexs_entries* \
    class attribute contains a dictionary with the regular expressions \
    for the fields in each record and a key name for each. Finally, the \
    *dpa_fixture_path* class attribute defines the path of the .json file \
    that contains a compressed dictionary with communes and constituencies.
    '''
    # regex_begin = r'^REPUBLICA\s+DE\s+CHILE'
    #: roll name regex
    regex_roll = r'PADRO?Ó?N\s+ELECTORAL\s+[A-Z,\s-]+\d+'
    #: region regex
    regex_region = r'REGIO?Ó?N[0,]*\s*:\s*([A-ZÑ\'\s.]*\s{3})'
    #: commune regex
    regex_commune = r'COMUNA[0,]*\s*:\s*([A-ZÑ\' ]*\s{3})'
    #: province regex
    regex_province = r'PROVINCIA[0,]*\s*:\s*([A-ZÑ\' ]*)'
    #: total entries regex (optional)
    regex_total_entries = r'Registros\s*:\s*(\d+)'
    #: pagination regex (optional)
    regex_pagination = r'[PAaáGgIiNn]+\s*:?\s*(\d*)\s*de\s*(\d*)'
    #: regex's for parsing entries.
    regexs_entries = {
        'name': r'^[A-ZÑa-z\s]+',
        'rut': r'\d*\.?\d+\.\d+-[0-9kK]',
        'sex': r'\s(VAR|MUJ)[ONER]*\s',
        'table': r'\s(\d+\s?\w?)\s*\d*$'}
    #: path to commune-circuns json.
    dpa_fixture_path = '../utils/DPA-commune-circuns.json'

    def run(self):
        '''
        Method that starts the voter registry sheet analyzer by executing:
        * :meth:`decompose <.RollParser.decompose>`
        * :meth:`parse_header <.RollParser.parse_header>`
        * :meth:`parse_fields <.RollParser.parse_fields>`
        * :meth:`parse_entries <.RollParser.parse_entries>`

        It measures the duration times of each method executed and saves \
        them in the :attr:`metadata[times] <.RollParser.metadata>` \
        property.
        '''
        self.decompose()
        header_at = dt.now()
        self.parse_header()
        fields_at = dt.now()
        self.parse_fields()
        entries_at = dt.now()
        self.parse_entries()
        finish_at = dt.now()
        times = {
            'header': fields_at - header_at,
            'fields': entries_at - fields_at,
            'entries': finish_at - entries_at}
        times['total'] = times['header'] + times['fields'] + times['entries']
        if 'times' not in self._metadata:
            self._metadata['times'] = {}
        self._metadata['times'].update(times)

    def decompose(self):
        '''
        Method that descompose a :attr:`sheet <.RollParser.sheet>` \
        of the electoral roll in a text string into a list with each line.
        '''
        if not self.is_decomposed:
            self._sheet = self._sheet.split('\n')

    @property
    def is_decomposed(self):
        '''
        :return: boolean.

        Property that indicates whether the '*sheet*' property is decomposed \
        into a list of lines or not.
        '''
        return isinstance(self.sheet, list)

    @property
    def sheet(self):
        '''
        Property that contains a text string of the entire sheet to parse or \
        the list of text lines if it is decomposed.

        >>> parser.is_descomposed
        False
        >>> parser.sheet
        'text-\\n-string'
        >>> parser.descompose
        >>> parser.sheet
        ['text-', '-string']
        '''
        return self._sheet

    def parse_header(self):
        '''
        Method that parses the head of the sheet and extracts the data \
        from *roll*, *election*, *year*, *region*, *province* and \
        *commune* to store it in the :attr:`header \
        <.RollParser.header>` property.

        It also builds a unique identifier of the electoral roll that \
        it stores in the :attr:`metadata <.RollParser.metadata>` \
        property with the *rid* key.
        '''
        def __identify(header):
            id_roll = re.sub(r'[0-9\tYa-z]', '', ''.join(
                [w[0] for w in header['roll'].split(' ') if len(w) > 2]))
            return f'{id_roll}-{str(header["year"])}'
        idx = self.__get_fields_index()
        target = ' \t '.join(self.sheet[:idx])
        attributes = ['roll', 'region', 'commune', 'province']
        header = {}
        for attr in attributes:
            header[attr] = self.__parser(
                regex=getattr(self, 'regex_'+attr),
                target=target, ecode='header-no-'+attr)
        year = self.__parser(
            r'\d+$', header['roll'], 'header-no-year')
        if year:
            header['year'] = int(year)
        total_entries = re.findall(self.regex_total_entries, target)
        if total_entries:
            header['total_entries'] = int(total_entries[0].strip())
        pagination = re.findall(self.regex_pagination, target)
        if pagination and len(pagination[0]) > 2 and pagination[0]:
            header['pagination'] = pagination[0]
        self._metadata['rid'] = __identify(header)
        self._header = header
        with open(os.path.dirname(__file__)+'/'+self.dpa_fixture_path) as f:
            fixture = json.load(f)
        if 'PAGINA' in header['commune']:
            header['commune'] = header['commune'].replace('PAGINA', '').strip()
        if header['commune'] in fixture:
            self._circuns = fixture[header['commune']]
        else:
            self._circuns = None
            self._errors.append({
                'code': 'commune-not-in-fixture',
                'fixture': self.dpa_fixture_path,
                'target': header['commune']})

    @property
    def header(self):
        '''
        Property that contains the result of method :meth:`parse_header \
        <.SheetRollParser.parse_header>`. It consists of a \
        dictionary with the data from the header of the electoral roll sheet.

        >>> parser.header
        {
            'roll': 'PADRON...',
            'election': 'ELECCION...',
            'year':     2020,
            'region':   'METRO...',
            'commune':  'SANTIA...',
            'province': 'SANTIA...',
        }
        '''
        return self._header

    def parse_fields(self):
        '''
        Method to analyze and extract the fields of the electoral roll. \
        The direct fields of the sheet (*nombre*, *c-identidad*, *sex|o*, \
        *domicilio-electoral*, *circunscripcion* y *mesa*) are taken and \
        commune, province and region (*comuna*, *provincia*, *region*) are \
        added. Result is stored in the :attr:`fields \
        <.RollParser.fields>` property, the method returns nothing.
        '''
        index = self.__get_fields_index()
        if not index:
            return None
        fields_line = self.sheet[index]
        fields_line = fields_line.replace(
            'LIO ELE', 'LIO-ELE')
        self._fields = list(map(
            lambda x: slugify(x),
            fields_line.split()))
        if self.more_fields:
            self._fields.insert(3, 'comuna')
            self._fields.insert(3, 'provincia')
            self._fields.insert(3, 'region')
            self._fields.append('reference')

    @property
    def fields(self):
        '''
        Property that contains the fields of the electoral roll \
        detected in the sheet through the :meth:`parse_fields \
        <.RollParser.parse_fields>` method.
        '''
        return self._fields

    def parse_entries(self):
        '''
        Method that analyzes and extracts each data entry from the voter \
        registration sheet.

        First determine if each line of text is well composed, that is, \
        it begins with at least one letter and ends with a number or a space \
        next to a single letter. Then each line of text is analyzed as if it \
        were an input through the :meth:`parse_entry \
        <.RollParser.parse_entry>` method.

        Afterwards, the lines considered malformed are internally processed, \
        joining them in relation to whether they start with a letter or a \
        space. Then use the :meth:`parse_entry \
        <.RollParser.parse_entry>` method again for each of them. \
        Those that are rescued will remain in the :attr:`metadata \
        <.RollParser.metadata>` property in the keys *entires* > \
        *rescue*.
        '''
        index = self.__get_fields_index() + 1
        if not index:
            return None
        entries = []
        malformed = []
        for line in self.sheet[index:]:
            if not re.match(r'^\w+.+\d+\s?\w?$', line):
                if line.strip():
                    malformed.append(line)
                continue
            entry = self.parse_entry(line)
            if entry:
                entries.append(entry)
        total_entries = len(entries)
        self.metadata['entries'] = {}
        if malformed:
            # try to take entry again
            rescue_entries = self.__rescue_entries(malformed)
            entries += rescue_entries
            self._metadata['entries']['rescue'] = len(entries) - total_entries
            total_entries = len(entries)
        self._metadata['entries']['errors'] = len(self.errors)
        self._metadata['entries']['total'] = total_entries
        self._entries = entries

    def parse_entry(self, line):
        '''
        A method that extracts the data from a voter registry entry in text \
        line format.

        Finds the fields found by regular expressions that are stored in the \
        class attribute: :attr: `regexs_entries \
        <.RollParser.regexs_entries>`.

        Then it looks for the district from a list according to its commune \
        and in relation to this it determines the place of the electoral \
        domicile.
        '''
        def __regex_fields(line, fields):
            attr_fields = ['name', 'rut', 'sex', 'table']
            fields = [fields[0], fields[1], fields[2], fields[-1]]
            values = []
            for idx, field in enumerate(fields):
                value = self.__parser(
                    regex=self.regexs_entries[attr_fields[idx]],
                    target=line, ecode='entry-'+field)
                values.append(value)
                if not value:
                    if field not in self._metadata['nulls']:
                        self._metadata['nulls'][field] = 1
                    else:
                        self._metadata['nulls'][field] += 1
            return values

        def __parse_circun(line, field):
            largest = len(sorted(self.circuns, key=lambda x: len(x))[-1])
            table_position = re.search(
                r'\s*'+self.regexs_entries['table'], line)
            if not table_position:
                table_position = 150
            else:
                table_position = table_position.start()
            cut = table_position - largest
            value = [c for c in self.circuns if c+'  ' in line[cut:]]
            if not value:
                value = [c for c in self.circuns if c in line[cut:]]
                if value and len(value) == 1:
                    return value[0]
                self._errors.append({
                    'code': 'entry-' + field + '-not-found',
                    'circuns': self.circuns,
                    'target': line})
                if field not in self._metadata['nulls']:
                    self._metadata['nulls'][field] = 1
                else:
                    self._metadata['nulls'][field] += 1
                return None
            return value[0]

        def __parse_dir(line, sex, circun, field):
            ini = re.search(sex, line)
            end = re.search(
                str(circun) + r'\s*' + self.regexs_entries['table'], line)
            if not end:
                end = -1
            else:
                end = end.start()
            if not ini:
                self._errors.append({
                    'code': 'entry-' + field + '-not-found',
                    'reason': 'sex end not found',
                    'target': line})
            else:
                ini = ini.end()
                direction = line[ini:end].strip()
                if direction:
                    return direction
            if field not in self._metadata['nulls']:
                self._metadata['nulls'][field] = 1
            else:
                self._metadata['nulls'][field] += 1
        if 'nulls' not in self.metadata:
            self._metadata['nulls'] = {'total': 0}
        regex_vals = __regex_fields(line, self.fields)
        entry = regex_vals
        circun = __parse_circun(line, self.fields[-2])
        entry.insert(3, circun)
        direction = __parse_dir(
            line, self.regexs_entries['sex'], circun, self.fields[-3])
        entry.insert(3, direction)
        if self.more_fields:
            entry.insert(3, self.header['commune'])
            entry.insert(3, self.header['province'])
            entry.insert(3, self.header['region'])
            reference = self.metadata['rid']
            reference += '-' + slugify(self.header['commune'])
            if 'pagination' in self.header:
                pag = self.header['pagination']
                reference += '-page-' + str(pag[0]) + '-of-' + str(pag[1])
            entry.append(reference)
        return entry

    def __rescue_entries(self, malformed):
        pentries = []
        entries = []
        lastentry = ''
        for txt in malformed:
            if re.match(r'^\w+', txt):
                if lastentry:
                    pentries.append(lastentry)
                lastentry = txt
            elif lastentry:
                lastentry += txt
            else:
                self._errors.append({
                    'code': 'malformed-no-start',
                    'target': txt})
        for e in pentries:
            entry = self.parse_entry(e)
            if not entry:
                self._errors.append({
                    'code': 'malformed-no-entry',
                    'target': e})
            else:
                entries.append(entry)
        return entries

    def __get_fields_index(self):
        if self.fields_index:
            return self.fields_index
        regex = r'^NOMBRE\s+C'
        for idx, line in enumerate(self.sheet):
            if re.match(regex, line):
                self._fields_index = idx
                return idx
        self._errors.append({
            'code': 'fields-no-index',
            'regex': regex,
            'target': self.sheet})

    def __parser(self, regex, target, ecode):
        parsed = re.findall(regex, target) if isinstance(target, str) else None
        if not bool(parsed):
            self._errors.append({
                'code': ecode,
                'regex': regex,
                'target': target})
            return None
        return re.sub(r'\s+', ' ', parsed[0].strip())

    @property
    def entries(self):
        '''
        Property containing a list of entries from the electoral roll \
        sheet. Each entry corresponds to a list of data in the order of the \
        fields defined in the :attr:`fields <RollParser.fields>` \
        property.
        '''
        return self._entries

    @property
    def metadata(self):
        '''
        Property contains the metadata extracted during the parser analysis \
        of the electoral roll sheet.

        The metadata is stored as a dictionary, the *rid* key \
        corresponding to the unique identifier of the voter registry of the \
        sheet, the *times* key stores how long the analysis took (in \
        total, during header, fields and entries), the *entries* key \
        contains the total number of entries extracted , the amount of \
        rescued and errors. Finally, the NULLS key contains the total \
        number of null data inside each row or entry, as well as the detail \
        of the fields, if there is any.

        >>> parser.metadata
        {
            'rid': 'PEA-EM-2016',
            'entries': {'total': 1, 'rescue': 0, 'errors': 0},
            'nulls': {'total': 0}
        }
        '''
        return self._metadata

    @property
    def errors(self):
        '''
        Property contains a list with the errors found in the sheet analysis. \
        Each error corresponds to a dictionary with at least two keys: \
        *code* with a semantic slug text of the error and '*target*' that \
        contains what generated the error.
        '''
        return self._errors

    @property
    def fields_index(self):
        '''
        Property that contains the index where the fields are located in the \
        decomposed sheet as a list.
        '''
        return self._fields_index

    @property
    def circuns(self):
        '''
        :return: a dictionary with communes as key and list of \
            circunscriptions as value.

        Property that contains the possible electoral circunscriptions \
        within the commune defined in the :attr:`header \
        <.RollParser.header>` property. It will return None if \
        the :meth:`parse_header <.RollParser.parse_header>` \
        method has not been executed.
        '''
        return self._circuns

    @property
    def more_fields(self):
        '''
        :return: boolean.

        Property where the option of whether to add fields to the \
        input is stored.
        '''
        return self._more_fields

    def __launch_props(self):
        self._fields_index = None
        self._metadata = {}
        self._header = {}
        self._fields = []
        self._entries = []
        self._errors = []

    def __init__(self, sheet, auto=True, more_fields=True, *args, **kwargs):
        if not isinstance(sheet, str) or not sheet:
            raise TypeError('\'sheet\' arg must be string')
        self._more_fields = bool(more_fields)
        self.__launch_props()
        self._sheet = sheet
        if auto:
            self.run()



from pathlib import Path
from datetime import datetime, timedelta
from string import ascii_letters
import random
import yaml
from slugify import slugify
import csv


class RollExporter:
    '''
    :class:`RollExporter <.RollExporter>` is a class for exporting electoral \
    roll data in csv files.

    :param bool export: If the export is activated (default False)
    :param str output: directory to store the data in .csv (see more in \
        :attr:`output <.RollExporter.output>`.
    :param str mode: determines the data export mode in files (*unified* o \
        *separated*, see more in :attr:`mode <.RollExporter.mode>`).
    :param str mode_sep: Criteria for separating files in export in \
        separate mode (*region* o *commune*, see more in :attr:`mode \
        <.RollExporter.mode>`).
    :param bool random_suffix: Determines whether exported files have a \
        random text string appended to the end.
    :param bool summary: Determines whether to generate a summary file of \
        the export and the extracted data.

    It is instantiated within an instance of :class:`ElectoralRoll \
    <.ElectoralRoll>`.
    '''

    #: Available export modes.
    modes = ['unified', 'separated']
    #: Available file separation modes
    mode_sep_opts = ['commune', 'region']

    def export_sheet(self, parsed):
        '''
        :param obj parsed: an instance of :class:`RollParser <.RollParser>`.
        :return: the absolute path of the file where the data was exported.

        :meth:`export_sheet <.RollExporter.export_sheet>` is a method of \
        exporting the data from a parsed sheet into files as configured \
        in the constructor.
        '''
        if not self.is_active:
            return None
        file, created = self.get_or_create_file(parsed)
        with file.open('a') as f:
            writer = csv.writer(f)
            if created:
                writer.writerow(parsed.fields)
            for entry in parsed.entries:
                writer.writerow(entry)
        return str(file.absolute())

    def export_summary(self, rid, metadata):
        '''
        :param str rid: identifier of the electoral roll
        :param dict metadata: metadata of the electoral roll.

        :meth:`export_summary <.RollExporter.export_summary>` is a \
        method that exports the metadata of the electoral roll as a \
        summary in a yaml file.
        '''
        def metadata_serializer(meta):
            meta = {**meta}
            for key in meta:
                if isinstance(meta[key], dict):
                    meta[key] = metadata_serializer(meta[key])
                elif isinstance(meta[key], datetime) or isinstance(
                        meta[key], timedelta):
                    meta[key] = str(meta[key])
                else:
                    continue
            return meta
        if not self.is_active or not self.summary:
            return None
        metadata = metadata_serializer(metadata)
        file = self.create_summary(rid)
        with file.open('w') as f:
            f.write('# Serveliza summary\n')
            f.write(yaml.dump(metadata))
        return str(file.absolute())

    def get_or_create_file(self, parsed):
        suffix, created = '', False
        if self.mode == 'separated':
            suffix = slugify(parsed.header[self.mode_sep]) + '-'
        name = f'{parsed.metadata["rid"]}-{suffix}data'
        if self.random_suffix:
            name += f'-{self.random_suffix}'
        name += '.csv'
        file = self.output / name
        if not file.exists():
            created = True
            file.touch()
        return file, created

    def create_summary(self, rid):
        name = f'{rid}-summary'
        if self.random_suffix:
            name += f'-{self.random_suffix}'
        name += '.txt'
        summary = self.output / name
        return summary

    @property
    def output(self):
        '''
        Directory to store the data in .csv.
        '''
        return self._output

    @output.setter
    def output(self, output):
        output_path = Path(str(output))
        try:
            output_path.mkdir(parents=True, exist_ok=True)
        except FileExistsError:
            raise FileExistsError(output+' file already exists.')
        self._output = output_path

    @property
    def mode(self):
        '''
        Determines the data export mode in files. If it is "unified" \
        (default) it creates a single csv file with the data, or if it \
        is "separated" into several according to communal or \
        regional criteria.'
        '''
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode not in self.modes:
            raise TypeError('mode must be: '
                            ','.join(self.modes))
        self._mode = mode

    @property
    def mode_sep(self):
        '''
        Criteria for separating files in export in separate mode.
        '''
        return self._mode_sep

    @mode_sep.setter
    def mode_sep(self, mode_sep):
        if mode_sep not in self.mode_sep_opts:
            raise TypeError('mode_sep must be: '
                            ','.join(self.mode_sep_opts))
        self._mode_sep = mode_sep

    @property
    def is_active(self):
        '''
        :return: boolean.

        Property that indicates if the memorizer is active as defined \
        in the constructor.
        '''
        return self._is_active

    @property
    def random_suffix(self):
        '''
        Determines whether exported files have a random \
        text string appended to the end.
        '''
        return self._random_suffix

    @property
    def summary(self):
        '''
        Determines whether to generate a summary file of \
        the export and the extracted data.
        '''
        return self._summary

    def __init__(self, *args, **kwargs):
        self._is_active = bool(kwargs.get('export', False))
        if not self.is_active:
            return None
        self.output = kwargs.get('output', 'output')
        self.mode = kwargs.get('mode', 'unified')
        if self.mode == 'separated':
            self.mode_sep = kwargs.get('mode_sep', 'commune')
        self._random_suffix = False
        if kwargs.get('random_suffix', True):
            self._random_suffix = ''.join([random.choice(
                ascii_letters) for x in range(5)])
        self._summary = bool(kwargs.get('summary', True))



from pathlib import Path
from datetime import datetime


def is_valid_pdf(pathfile, raise_exception=False):
    '''
    '''
    path = Path(str(pathfile))
    msg, valid = '', False
    if not path.exists():
        msg = 'Path no exists.'
    elif not path.is_file():
        msg = 'Path is not a file.'
    elif path.suffix not in ['.pdf', '.PDF']:
        msg = 'Path dont have pdf suffix.'
    else:  # valid pdf
        valid = True
    if not valid and raise_exception:
        raise TypeError(msg)
    return valid


def get_all_pdf_in_path(path, recursively=False):
    '''
    '''
    pattern = '**/*' if recursively else '*'
    return [x for x in Path(path).glob(pattern) if x.is_file()
            and x.suffix in ['.pdf', '.PDF']]


def get_metadata_from_pdfs(filelist, output='dict'):
    '''
    '''
    if not isinstance(filelist, list) or output not in ['dict', 'list']:
        TypeError('filelist must be list and output '
                  'must be \'dict\' or \'list\'')
    metadata = []
    for pdf in filelist:
        path = Path(pdf)
        meta = {
            'name': path.name, 'bytes': path.stat().st_size,
            'relative': str(path.relative_to('.')),
            'absolute': str(path.absolute()),
            'mtime': datetime.fromtimestamp(path.stat().st_mtime),
            'atime': datetime.fromtimestamp(path.stat().st_mtime),
            }
        metadata.append(meta)
    if output == 'list':
        return metadata
    elif output == 'dict':
        dict_metadata = {}
        for meta in metadata:
            dict_metadata[meta['name']] = meta
        return dict_metadata
